@@ -31,14 +31,15 @@ from impala.interface import Connection, Cursor, _bind_parameters
 from impala.error import (NotSupportedError, OperationalError,
                           ProgrammingError, HiveServer2Error)
 from impala._thrift_api import (
-    get_socket, get_transport, TTransportException, TBinaryProtocol,
-    TOpenSessionReq, TFetchResultsReq, TCloseSessionReq, TExecuteStatementReq,
-    TGetInfoReq, TGetInfoType, TTypeId, TFetchOrientation,
-    TGetResultSetMetadataReq, TStatusCode, TGetColumnsReq, TGetSchemasReq,
-    TGetTablesReq, TGetFunctionsReq, TGetOperationStatusReq, TOperationState,
-    TCancelOperationReq, TCloseOperationReq, TGetLogReq, TProtocolVersion,
-    TGetRuntimeProfileReq, TGetExecSummaryReq, ImpalaHiveServer2Service,
-    TExecStats, ThriftClient, TApplicationException)
+    get_socket, get_http_transport, get_transport, THttpClient,
+    TTransportException, TBinaryProtocol, TOpenSessionReq, TFetchResultsReq,
+    TCloseSessionReq, TExecuteStatementReq, TGetInfoReq, TGetInfoType, TTypeId,
+    TFetchOrientation, TGetResultSetMetadataReq, TStatusCode, TGetColumnsReq,
+    TGetSchemasReq, TGetTablesReq, TGetFunctionsReq, TGetOperationStatusReq,
+    TOperationState, TCancelOperationReq, TCloseOperationReq, TGetLogReq,
+    TProtocolVersion, TGetRuntimeProfileReq, TRuntimeProfileFormat,
+    TGetExecSummaryReq, ImpalaHiveServer2Service, TExecStats, ThriftClient,
+    TApplicationException)
 
 
 log = get_logger_and_init_null(__name__)
@@ -92,8 +93,6 @@ class HiveServer2Connection(Connection):
             expensive.) Only applies when using HS2 protocol versions > 6.
         dictify : bool, optional
             When `True` cursor will return key value pairs instead of rows.
-        batch_cursor : bool, optional
-            When `True` cursor will return CBatches directly rather than rows.
         fetch_error : bool, optional
             In versions of impala prior to 2.7.0, when an operation fails and
             the impalad returns an error state, the error message is not always
@@ -168,6 +167,8 @@ class HiveServer2Cursor(Cursor):
         self._closed = False
 
     def __del__(self):
+        if self._closed:
+            return
         try:
            self.close_operation()
         except Exception:
@@ -192,6 +193,26 @@ class HiveServer2Cursor(Cursor):
         # PEP 249
         return self._rowcount
 
+    @property
+    def rowcounts(self):
+        # Work around to get the number of rows modified for Inserts/Update/Delte statements
+        modifiedRows, errorRows = -1, -1
+        if self._last_operation_active:
+            logList = self.get_profile().split('\n')
+            resultDict = {}
+            subs = ['NumModifiedRows', 'NumRowErrors']
+            resultSet = [s for s in logList if any(item in s for item in subs)]
+            if resultSet:
+                for items in resultSet:
+                    key, value = items.split(':')
+                    key, value = key.strip(), value.strip()
+                    resultDict[key] = value
+
+                modifiedRows = int(resultDict.get('NumModifiedRows', -1))
+                errorRows = int(resultDict.get('NumRowErrors', -1))
+
+        return (modifiedRows, errorRows)      
+      
     @property
     def lastrowid(self):
         # PEP 249
@@ -259,11 +280,12 @@ class HiveServer2Cursor(Cursor):
         if exc_info:
             six.reraise(*exc_info)
 
-    def cancel_operation(self):
+    def cancel_operation(self, reset_state=True):
         if self._last_operation_active:
             log.info('Canceling active operation')
             self._last_operation.cancel()
-            self._reset_state()
+            if reset_state:
+                self._reset_state()
 
     def close_operation(self):
         if self._last_operation_active:
@@ -333,10 +355,15 @@ class HiveServer2Cursor(Cursor):
         """
         log.debug('Executing query %s', operation)
 
+        paramstyle = None
+        if configuration:
+            paramstyle = configuration.pop('paramstyle', None)
+
         def op():
             if parameters:
                 self._last_operation_string = _bind_parameters(operation,
-                                                               parameters)
+                                                               parameters,
+                                                               paramstyle)
             else:
                 self._last_operation_string = operation
 
@@ -394,6 +421,8 @@ class HiveServer2Cursor(Cursor):
             time.sleep(self._get_sleep_interval(loop_start))
 
     def status(self):
+        if self._last_operation is None:
+            raise ProgrammingError("Operation state is not available")
         return self._last_operation.get_status()
 
     def execution_failed(self):
@@ -429,11 +458,11 @@ class HiveServer2Cursor(Cursor):
             return 0.5
         return 1.0
 
-    def executemany(self, operation, seq_of_parameters):
+    def executemany(self, operation, seq_of_parameters, configuration=None):
         # PEP 249
         log.debug('Attempting to execute %s queries', len(seq_of_parameters))
         for parameters in seq_of_parameters:
-            self.execute(operation, parameters)
+            self.execute(operation, parameters, configuration)
             if self.has_result_set:
                 raise ProgrammingError("Operations that have result sets are "
                                        "not allowed with executemany.")
@@ -451,8 +480,10 @@ class HiveServer2Cursor(Cursor):
 
 
     def fetchcbatch(self):
-        '''Return a CBatch object of any data currently in the buffer or
-           if no data currently in buffer then fetch a batch'''
+        '''Return a CBatch object containing the next rows to be fetched. If data is
+           currently buffered, returns that data, otherwise fetches the next batch.
+           Returns None if no more rows are currently available. Note that if None
+           is returned, more rows may still be available in future.'''
         if not self._last_operation.is_columnar:
             raise NotSupportedError("Server does not support columnar "
                                     "fetching")
@@ -537,26 +568,29 @@ class HiveServer2Cursor(Cursor):
         return self.__next__()
 
     def __next__(self):
-        if not self.has_result_set:
-            raise ProgrammingError(
-                "Trying to fetch results on an operation with no results.")
-        if len(self._buffer) > 0:
-            log.debug('__next__: popping row out of buffer')
-            return self._buffer.pop()
-        elif self._last_operation_active:
-            log.debug('__next__: buffer empty and op is active => fetching '
-                      'more data')
-            self._buffer = self._last_operation.fetch(self.description,
-                                                      self.buffersize,
-                                                      convert_types=self.convert_types)
-            if len(self._buffer) == 0:
-                log.debug('__next__: no more data to fetch')
+        while True:
+            if not self.has_result_set:
+                raise ProgrammingError(
+                    "Trying to fetch results on an operation with no results.")
+            if len(self._buffer) > 0:
+                log.debug('__next__: popping row out of buffer')
+                return self._buffer.pop()
+            elif self._last_operation_active:
+                log.debug('__next__: buffer empty and op is active => fetching '
+                          'more data')
+                self._buffer = self._last_operation.fetch(self.description,
+                                                          self.buffersize,
+                                                          convert_types=self.convert_types)
+                if len(self._buffer) > 0:
+                  log.debug('__next__: popping row out of buffer')
+                  return self._buffer.pop()
+                if not self._buffer.expect_more_rows:
+                    log.debug('__next__: no more data to fetch')
+                    raise StopIteration
+                # If we didn't get rows, but more are expected, need to iterate again.
+            else:
+                log.debug('__next__: buffer empty')
                 raise StopIteration
-            log.debug('__next__: popping row out of buffer')
-            return self._buffer.pop()
-        else:
-            log.debug('__next__: buffer empty')
-            raise StopIteration
 
     def ping(self):
         """Checks connection to server by requesting some info."""
@@ -564,10 +598,14 @@ class HiveServer2Cursor(Cursor):
         return self.session.ping()
 
     def get_log(self):
+        if self._last_operation is None:
+            raise ProgrammingError("Operation state is not available")
         return self._last_operation.get_log()
 
-    def get_profile(self):
-        return self._last_operation.get_profile()
+    def get_profile(self, profile_format=TRuntimeProfileFormat.STRING):
+        if self._last_operation is None:
+            raise ProgrammingError("Operation state is not available")
+        return self._last_operation.get_profile(profile_format=profile_format)
 
     def get_summary(self):
         return self._last_operation.get_summary()
@@ -744,23 +782,46 @@ def threaded(func):
 
 def connect(host, port, timeout=None, use_ssl=False, ca_cert=None,
             user=None, password=None, kerberos_service_name='impala',
-            auth_mechanism=None):
+            auth_mechanism=None, krb_host=None, use_http_transport=False,
+            http_path=''):
     log.debug('Connecting to HiveServer2 %s:%s with %s authentication '
-             'mechanism', host, port, auth_mechanism)
-    sock = get_socket(host, port, use_ssl, ca_cert)
-    if timeout is not None:
-        timeout = timeout * 1000.  # TSocket expects millis
-    if six.PY2:
-        sock.setTimeout(timeout)
-    elif six.PY3:
-        try:
-            # thriftpy has a release where set_timeout is missing
-            sock.set_timeout(timeout)
-        except AttributeError:
-            sock.socket_timeout = timeout
-            sock.connect_timeout = timeout
-    transport = get_transport(sock, host, kerberos_service_name,
-                              auth_mechanism, user, password)
+              'mechanism', host, port, auth_mechanism)
+
+    if use_http_transport:
+        # TODO(#362): Add server authentication with thrift 0.12.
+        if ca_cert:
+            raise NotSupportedError("Server authentication is not supported " +
+                                    "with HTTP endpoints")
+        if krb_host:
+            raise NotSupportedError("Kerberos authentication is not " +
+                                    "supported with HTTP endpoints")
+        transport = get_http_transport(host, port, http_path=http_path,
+                                       use_ssl=use_ssl, ca_cert=ca_cert,
+                                       user=user, password=password,
+                                       auth_mechanism=auth_mechanism)
+    else:
+        sock = get_socket(host, port, use_ssl, ca_cert)
+
+        if krb_host:
+            kerberos_host = krb_host
+        else:
+            kerberos_host = host
+
+        if timeout is not None:
+            timeout = timeout * 1000.  # TSocket expects millis
+        if six.PY2:
+            sock.setTimeout(timeout)
+        elif six.PY3:
+            try:
+                # thriftpy has a release where set_timeout is missing
+                sock.set_timeout(timeout)
+            except AttributeError:
+                sock.socket_timeout = timeout
+                sock.connect_timeout = timeout
+        log.debug('sock=%s', sock)
+        transport = get_transport(sock, kerberos_host, kerberos_service_name,
+                                auth_mechanism, user, password)
+
     transport.open()
     protocol = TBinaryProtocol(transport)
     if six.PY2:
@@ -769,8 +830,8 @@ def connect(host, port, timeout=None, use_ssl=False, ca_cert=None,
     elif six.PY3:
         # ThriftClient == TClient
         service = ThriftClient(ImpalaHiveServer2Service, protocol)
-    log.debug('sock=%s transport=%s protocol=%s service=%s', sock, transport,
-              protocol, service)
+    log.debug('transport=%s protocol=%s service=%s', transport, protocol,
+              service)
 
     return HS2Service(service)
 
@@ -837,7 +898,8 @@ class Column(object):
 
 class CBatch(Batch):
 
-    def __init__(self, trowset, schema, convert_types=True):
+    def __init__(self, trowset, expect_more_rows, schema, convert_types=True):
+        self.expect_more_rows = expect_more_rows
         self.schema = schema
         tcols = [_TTypeId_to_TColumnValue_getters[schema[i][1]](col)
                  for (i, col) in enumerate(trowset.columns)]
@@ -894,8 +956,9 @@ class CBatch(Batch):
 
 
 class RBatch(Batch):
-    def __init__(self, trowset, schema):
+    def __init__(self, trowset, expect_more_rows, schema):
         log.debug('RBatch: input TRowSet: %s', trowset)
+        self.expect_more_rows = expect_more_rows
         self.schema = schema
         self.rows = []
         for trow in trowset.rows:
@@ -1118,6 +1181,10 @@ class Operation(ThriftRPC):
         resp = self._rpc('GetOperationStatus', req)
         return TOperationState._VALUES_TO_NAMES[resp.operationState]
 
+    def get_state(self):
+        req = TGetOperationStatusReq(operationHandle=self.handle)
+        return self._rpc('GetOperationStatus', req)
+
     def get_log(self, max_rows=1024, orientation=TFetchOrientation.FETCH_NEXT):
         try:
             req = TGetLogReq(operationHandle=self.handle)
@@ -1137,16 +1204,19 @@ class Operation(ThriftRPC):
 
     def cancel(self):
         req = TCancelOperationReq(operationHandle=self.handle)
-        self._rpc('CancelOperation', req)
+        return self._rpc('CancelOperation', req)
 
     def close(self):
         req = TCloseOperationReq(operationHandle=self.handle)
-        self._rpc('CloseOperation', req)
+        return self._rpc('CloseOperation', req)
 
-    def get_profile(self):
+    def get_profile(self, profile_format=TRuntimeProfileFormat.STRING):
         req = TGetRuntimeProfileReq(operationHandle=self.handle,
-                                    sessionHandle=self.session.handle)
+                                    sessionHandle=self.session.handle,
+                                    format=profile_format)
         resp = self._rpc('GetRuntimeProfile', req)
+        if profile_format == TRuntimeProfileFormat.THRIFT:
+            return resp.thrift_profile
         return resp.profile
 
     def get_summary(self):
@@ -1170,16 +1240,16 @@ class Operation(ThriftRPC):
                                orientation=orientation,
                                maxRows=max_rows)
         resp = self._rpc('FetchResults', req)
-        return self._wrap_results(resp.results, schema,
+        return self._wrap_results(resp.results, resp.hasMoreRows, schema,
                                   convert_types=convert_types)
 
-    def _wrap_results(self, results, schema, convert_types=True):
+    def _wrap_results(self, results, expect_more_rows, schema, convert_types=True):
         if self.is_columnar:
             log.debug('fetch_results: constructing CBatch')
-            return CBatch(results, schema, convert_types=convert_types)
+            return CBatch(results, expect_more_rows, schema, convert_types=convert_types)
         else:
             log.debug('fetch_results: constructing RBatch')
-            return RBatch(results, schema)
+            return RBatch(results, expect_more_rows, schema)
 
     @property
     def is_columnar(self):
